# -*- coding: utf-8 -*-
#
# Copyright: (c) 2019, F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import os
import json
import pytest
import sys

if sys.version_info < (2, 7):
    pytestmark = pytest.mark.skip("F5 Ansible modules require Python >= 2.7")

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.f5networks.f5_modules.plugins.modules.bigip_message_routing_peer import (
    ApiParameters, ModuleParameters, ModuleManager, GenericModuleManager, ArgumentSpec
)
from ansible_collections.f5networks.f5_modules.tests.unit.compat import unittest
from ansible_collections.f5networks.f5_modules.tests.unit.compat.mock import Mock, patch
from ansible_collections.f5networks.f5_modules.tests.unit.modules.utils import set_module_args


fixture_path = os.path.join(os.path.dirname(__file__), 'fixtures')
fixture_data = {}


def load_fixture(name):
    path = os.path.join(fixture_path, name)

    if path in fixture_data:
        return fixture_data[path]

    with open(path) as f:
        data = f.read()

    try:
        data = json.loads(data)
    except Exception:
        pass

    fixture_data[path] = data
    return data


class TestParameters(unittest.TestCase):
    def test_module_parameters(self):
        args = dict(
            name='foo',
            partition='foobar',
            description='my description',
            auto_init='yes',
            auto_init_interval=1234,
            connection_mode='per-peer',
            number_of_connections=20,
            pool='/Common/example',
            ratio=10,
            transport_config='/Common/virtual'
        )

        p = ModuleParameters(params=args)
        assert p.name == 'foo'
        assert p.partition == 'foobar'
        assert p.description == 'my description'
        assert p.auto_init == 'enabled'
        assert p.auto_init_interval == 1234
        assert p.connection_mode == 'per-peer'
        assert p.number_of_connections == 20
        assert p.pool == '/Common/example'
        assert p.ratio == 10
        assert p.transport_config == '/Common/virtual'

    def test_api_parameters(self):
        args = load_fixture('load_generic_peer.json')

        p = ApiParameters(params=args)
        assert p.name == 'test'
        assert p.partition == 'Common'
        assert p.description == 'foobar'
        assert p.auto_init == 'disabled'
        assert p.auto_init_interval == 5000
        assert p.connection_mode == 'per-peer'
        assert p.number_of_connections == 1
        assert p.pool == '/Common/example'
        assert p.ratio == 1
        assert p.transport_config == '/Common/test_tranport'


class TestManager(unittest.TestCase):
    def setUp(self):
        self.spec = ArgumentSpec()
        self.p2 = patch('ansible_collections.f5networks.f5_modules.plugins.modules.bigip_message_routing_peer.tmos_version')
        self.p3 = patch('ansible_collections.f5networks.f5_modules.plugins.modules.bigip_message_routing_peer.send_teem')
        self.m2 = self.p2.start()
        self.m2.return_value = '14.1.0'
        self.m3 = self.p3.start()
        self.m3.return_value = True

    def tearDown(self):
        self.p2.stop()
        self.p3.stop()

    def test_create_generic_peer(self, *args):
        set_module_args(dict(
            name='foo',
            partition='foobar',
            description='my description',
            auto_init='yes',
            auto_init_interval=1234,
            connection_mode='per-peer',
            number_of_connections=20,
            pool='/Common/example',
            ratio=10,
            transport_config='/Common/virtual',
            provider=dict(
                server='localhost',
                password='password',
                user='admin'
            )
        ))

        module = AnsibleModule(
            argument_spec=self.spec.argument_spec,
            supports_check_mode=self.spec.supports_check_mode,
            required_if=self.spec.required_if
        )

        # Override methods in the specific type of manager
        gm = GenericModuleManager(module=module)
        gm.exists = Mock(return_value=False)
        gm.create_on_device = Mock(return_value=True)

        mm = ModuleManager(module=module)
        mm.version_less_than_14 = Mock(return_value=False)
        mm.get_manager = Mock(return_value=gm)

        results = mm.exec_module()

        assert results['changed'] is True
        assert results['auto_init'] == 'yes'
        assert results['description'] == 'my description'
        assert results['auto_init_interval'] == 1234
        assert results['connection_mode'] == 'per-peer'
        assert results['number_of_connections'] == 20
        assert results['pool'] == '/Common/example'
        assert results['ratio'] == 10
        assert results['transport_config'] == '/Common/virtual'

    def test_update_generic_peer(self, *args):
        set_module_args(dict(
            name='test',
            auto_init_interval=3000,
            ratio=120,
            provider=dict(
                server='localhost',
                password='password',
                user='admin'
            )
        ))

        current = ApiParameters(params=load_fixture('load_generic_peer.json'))

        module = AnsibleModule(
            argument_spec=self.spec.argument_spec,
            supports_check_mode=self.spec.supports_check_mode,
            required_if=self.spec.required_if
        )

        # Override methods in the specific type of manager
        gm = GenericModuleManager(module=module)
        gm.exists = Mock(return_value=True)
        gm.update_on_device = Mock(return_value=True)
        gm.read_current_from_device = Mock(return_value=current)

        mm = ModuleManager(module=module)
        mm.version_less_than_14 = Mock(return_value=False)
        mm.get_manager = Mock(return_value=gm)

        results = mm.exec_module()

        assert results['changed'] is True
        assert results['auto_init_interval'] == 3000
        assert results['ratio'] == 120
