#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
# Copyright 2019-2020 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__metaclass__ = type

ANSIBLE_METADATA = {'status': ['preview'],
                    'supported_by': 'community',
                    'metadata_version': '1.1'}

DOCUMENTATION = '''
---
module: fortios_system_modem
short_description: Configure MODEM in Fortinet's FortiOS and FortiGate.
description:
    - This module is able to configure a FortiGate or FortiOS (FOS) device by allowing the
      user to set and modify system feature and modem category.
      Examples include all parameters and values need to be adjusted to datasources before usage.
      Tested with FOS v6.0.0
version_added: "2.10"
author:
    - Link Zheng (@chillancezen)
    - Hongbin Lu (@fgtdev-hblu)
    - Frank Shen (@frankshen01)
    - Jie Xue (@JieX19)
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Legacy fortiosapi has been deprecated, httpapi is the preferred way to run playbooks
requirements:
    - ansible>=2.9.0
options:
    host:
        description:
            - FortiOS or FortiGate IP address.
        type: str
        required: false
    username:
        description:
            - FortiOS or FortiGate username.
        type: str
        required: false
    password:
        description:
            - FortiOS or FortiGate password.
        type: str
        default: ""
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        type: str
        default: root
    https:
        description:
            - Indicates if the requests towards FortiGate must use HTTPS protocol.
        type: bool
        default: true
    ssl_verify:
        description:
            - Ensures FortiGate certificate must be verified by a proper CA.
        type: bool
        default: true
    system_modem:
        description:
            - Configure MODEM.
        default: null
        type: dict
        suboptions:
            action:
                description:
                    - Dial up/stop MODEM.
                type: str
                choices:
                    - dial
                    - stop
                    - none
            altmode:
                description:
                    - Enable/disable altmode for installations using PPP in China.
                type: str
                choices:
                    - enable
                    - disable
            authtype1:
                description:
                    - Allowed authentication types for ISP 1.
                type: str
                choices:
                    - pap
                    - chap
                    - mschap
                    - mschapv2
            authtype2:
                description:
                    - Allowed authentication types for ISP 2.
                type: str
                choices:
                    - pap
                    - chap
                    - mschap
                    - mschapv2
            authtype3:
                description:
                    - Allowed authentication types for ISP 3.
                type: str
                choices:
                    - pap
                    - chap
                    - mschap
                    - mschapv2
            auto_dial:
                description:
                    - Enable/disable auto-dial after a reboot or disconnection.
                type: str
                choices:
                    - enable
                    - disable
            connect_timeout:
                description:
                    - Connection completion timeout (30 - 255 sec).
                type: int
            dial_cmd1:
                description:
                    - Dial command (this is often an ATD or ATDT command).
                type: str
            dial_cmd2:
                description:
                    - Dial command (this is often an ATD or ATDT command).
                type: str
            dial_cmd3:
                description:
                    - Dial command (this is often an ATD or ATDT command).
                type: str
            dial_on_demand:
                description:
                    - Enable/disable to dial the modem when packets are routed to the modem interface.
                type: str
                choices:
                    - enable
                    - disable
            distance:
                description:
                    - Distance of learned routes (1 - 255).
                type: int
            dont_send_CR1:
                description:
                    - Do not send CR when connected (ISP1).
                type: str
                choices:
                    - enable
                    - disable
            dont_send_CR2:
                description:
                    - Do not send CR when connected (ISP2).
                type: str
                choices:
                    - enable
                    - disable
            dont_send_CR3:
                description:
                    - Do not send CR when connected (ISP3).
                type: str
                choices:
                    - enable
                    - disable
            extra_init1:
                description:
                    - Extra initialization string to ISP 1.
                type: str
            extra_init2:
                description:
                    - Extra initialization string to ISP 2.
                type: str
            extra_init3:
                description:
                    - Extra initialization string to ISP 3.
                type: str
            holddown_timer:
                description:
                    - Hold down timer in seconds (1 - 60 sec).
                type: int
            idle_timer:
                description:
                    - MODEM connection idle time (1 - 9999 min).
                type: int
            interface:
                description:
                    - Name of redundant interface. Source system.interface.name.
                type: str
            lockdown_lac:
                description:
                    - Allow connection only to the specified Location Area Code (LAC).
                type: str
            mode:
                description:
                    - Set MODEM operation mode to redundant or standalone.
                type: str
                choices:
                    - standalone
                    - redundant
            network_init:
                description:
                    - AT command to set the Network name/type (AT+COPS=<mode>,[<format>,<oper>[,<AcT>]]).
                type: str
            passwd1:
                description:
                    - Password to access the specified dialup account.
                type: str
            passwd2:
                description:
                    - Password to access the specified dialup account.
                type: str
            passwd3:
                description:
                    - Password to access the specified dialup account.
                type: str
            peer_modem1:
                description:
                    - Specify peer MODEM type for phone1.
                type: str
                choices:
                    - generic
                    - actiontec
                    - ascend_TNT
            peer_modem2:
                description:
                    - Specify peer MODEM type for phone2.
                type: str
                choices:
                    - generic
                    - actiontec
                    - ascend_TNT
            peer_modem3:
                description:
                    - Specify peer MODEM type for phone3.
                type: str
                choices:
                    - generic
                    - actiontec
                    - ascend_TNT
            phone1:
                description:
                    - Phone number to connect to the dialup account (must not contain spaces, and should include standard special characters).
                type: str
            phone2:
                description:
                    - Phone number to connect to the dialup account (must not contain spaces, and should include standard special characters).
                type: str
            phone3:
                description:
                    - Phone number to connect to the dialup account (must not contain spaces, and should include standard special characters).
                type: str
            pin_init:
                description:
                    - AT command to set the PIN (AT+PIN=<pin>).
                type: str
            ppp_echo_request1:
                description:
                    - Enable/disable PPP echo-request to ISP 1.
                type: str
                choices:
                    - enable
                    - disable
            ppp_echo_request2:
                description:
                    - Enable/disable PPP echo-request to ISP 2.
                type: str
                choices:
                    - enable
                    - disable
            ppp_echo_request3:
                description:
                    - Enable/disable PPP echo-request to ISP 3.
                type: str
                choices:
                    - enable
                    - disable
            priority:
                description:
                    - Priority of learned routes (0 - 4294967295).
                type: int
            redial:
                description:
                    - Redial limit (1 - 10 attempts, none = redial forever).
                type: str
                choices:
                    - none
                    - 1
                    - 2
                    - 3
                    - 4
                    - 5
                    - 6
                    - 7
                    - 8
                    - 9
                    - 10
            reset:
                description:
                    - Number of dial attempts before resetting modem (0 = never reset).
                type: int
            status:
                description:
                    - Enable/disable Modem support (equivalent to bringing an interface up or down).
                type: str
                choices:
                    - enable
                    - disable
            traffic_check:
                description:
                    - Enable/disable traffic-check.
                type: str
                choices:
                    - enable
                    - disable
            username1:
                description:
                    - User name to access the specified dialup account.
                type: str
            username2:
                description:
                    - User name to access the specified dialup account.
                type: str
            username3:
                description:
                    - User name to access the specified dialup account.
                type: str
            wireless_port:
                description:
                    - Enter wireless port number, 0 for default, 1 for first port, ... (0 - 4294967295)
                type: int
'''

EXAMPLES = '''
- hosts: fortigates
  collections:
    - fortinet.fortios
  connection: httpapi
  vars:
   vdom: "root"
   ansible_httpapi_use_ssl: yes
   ansible_httpapi_validate_certs: no
   ansible_httpapi_port: 443
  tasks:
  - name: Configure MODEM.
    fortios_system_modem:
      vdom:  "{{ vdom }}"
      system_modem:
        action: "dial"
        altmode: "enable"
        authtype1: "pap"
        authtype2: "pap"
        authtype3: "pap"
        auto_dial: "enable"
        connect_timeout: "9"
        dial_cmd1: "<your_own_value>"
        dial_cmd2: "<your_own_value>"
        dial_cmd3: "<your_own_value>"
        dial_on_demand: "enable"
        distance: "14"
        dont_send_CR1: "enable"
        dont_send_CR2: "enable"
        dont_send_CR3: "enable"
        extra_init1: "<your_own_value>"
        extra_init2: "<your_own_value>"
        extra_init3: "<your_own_value>"
        holddown_timer: "21"
        idle_timer: "22"
        interface: "<your_own_value> (source system.interface.name)"
        lockdown_lac: "<your_own_value>"
        mode: "standalone"
        network_init: "<your_own_value>"
        passwd1: "<your_own_value>"
        passwd2: "<your_own_value>"
        passwd3: "<your_own_value>"
        peer_modem1: "generic"
        peer_modem2: "generic"
        peer_modem3: "generic"
        phone1: "<your_own_value>"
        phone2: "<your_own_value>"
        phone3: "<your_own_value>"
        pin_init: "<your_own_value>"
        ppp_echo_request1: "enable"
        ppp_echo_request2: "enable"
        ppp_echo_request3: "enable"
        priority: "40"
        redial: "none"
        reset: "42"
        status: "enable"
        traffic_check: "enable"
        username1: "<your_own_value>"
        username2: "<your_own_value>"
        username3: "<your_own_value>"
        wireless_port: "48"
'''

RETURN = '''
build:
  description: Build number of the fortigate image
  returned: always
  type: str
  sample: '1547'
http_method:
  description: Last method used to provision the content into FortiGate
  returned: always
  type: str
  sample: 'PUT'
http_status:
  description: Last result given by FortiGate on last operation applied
  returned: always
  type: str
  sample: "200"
mkey:
  description: Master key (id) used in the last call to FortiGate
  returned: success
  type: str
  sample: "id"
name:
  description: Name of the table used to fulfill the request
  returned: always
  type: str
  sample: "urlfilter"
path:
  description: Path of the table used to fulfill the request
  returned: always
  type: str
  sample: "webfilter"
revision:
  description: Internal revision number
  returned: always
  type: str
  sample: "17.0.2.10658"
serial:
  description: Serial number of the unit
  returned: always
  type: str
  sample: "FGVMEVYYQT3AB5352"
status:
  description: Indication of the operation's result
  returned: always
  type: str
  sample: "success"
vdom:
  description: Virtual domain used
  returned: always
  type: str
  sample: "root"
version:
  description: Version of the FortiGate
  returned: always
  type: str
  sample: "v5.6.3"

'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.connection import Connection
from ansible_collections.fortinet.fortios.plugins.module_utils.fortios.fortios import FortiOSHandler
from ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.common import FAIL_SOCKET_MSG


def login(data, fos):
    host = data['host']
    username = data['username']
    password = data['password']
    ssl_verify = data['ssl_verify']

    fos.debug('on')
    if 'https' in data and not data['https']:
        fos.https('off')
    else:
        fos.https('on')

    fos.login(host, username, password, verify=ssl_verify)


def filter_system_modem_data(json):
    option_list = ['action', 'altmode', 'authtype1',
                   'authtype2', 'authtype3', 'auto_dial',
                   'connect_timeout', 'dial_cmd1', 'dial_cmd2',
                   'dial_cmd3', 'dial_on_demand', 'distance',
                   'dont_send_CR1', 'dont_send_CR2', 'dont_send_CR3',
                   'extra_init1', 'extra_init2', 'extra_init3',
                   'holddown_timer', 'idle_timer', 'interface',
                   'lockdown_lac', 'mode', 'network_init',
                   'passwd1', 'passwd2', 'passwd3',
                   'peer_modem1', 'peer_modem2', 'peer_modem3',
                   'phone1', 'phone2', 'phone3',
                   'pin_init', 'ppp_echo_request1', 'ppp_echo_request2',
                   'ppp_echo_request3', 'priority', 'redial',
                   'reset', 'status', 'traffic_check',
                   'username1', 'username2', 'username3',
                   'wireless_port']
    dictionary = {}

    for attribute in option_list:
        if attribute in json and json[attribute] is not None:
            dictionary[attribute] = json[attribute]

    return dictionary


def underscore_to_hyphen(data):
    if isinstance(data, list):
        for i, elem in enumerate(data):
            data[i] = underscore_to_hyphen(elem)
    elif isinstance(data, dict):
        new_data = {}
        for k, v in data.items():
            new_data[k.replace('_', '-')] = underscore_to_hyphen(v)
        data = new_data

    return data


def system_modem(data, fos):
    vdom = data['vdom']
    system_modem_data = data['system_modem']
    filtered_data = underscore_to_hyphen(filter_system_modem_data(system_modem_data))

    return fos.set('system',
                   'modem',
                   data=filtered_data,
                   vdom=vdom)


def is_successful_status(status):
    return status['status'] == "success" or \
        status['http_method'] == "DELETE" and status['http_status'] == 404


def fortios_system(data, fos):

    if data['system_modem']:
        resp = system_modem(data, fos)

    return not is_successful_status(resp), \
        resp['status'] == "success" and \
        (resp['revision_changed'] if 'revision_changed' in resp else True), \
        resp


def main():
    fields = {
        "host": {"required": False, "type": "str"},
        "username": {"required": False, "type": "str"},
        "password": {"required": False, "type": "str", "default": "", "no_log": True},
        "vdom": {"required": False, "type": "str", "default": "root"},
        "https": {"required": False, "type": "bool", "default": True},
        "ssl_verify": {"required": False, "type": "bool", "default": True},
        "system_modem": {
            "required": False, "type": "dict", "default": None,
            "options": {
                "action": {"required": False, "type": "str",
                           "choices": ["dial",
                                       "stop",
                                       "none"]},
                "altmode": {"required": False, "type": "str",
                            "choices": ["enable",
                                        "disable"]},
                "authtype1": {"required": False, "type": "str",
                              "choices": ["pap",
                                          "chap",
                                          "mschap",
                                          "mschapv2"]},
                "authtype2": {"required": False, "type": "str",
                              "choices": ["pap",
                                          "chap",
                                          "mschap",
                                          "mschapv2"]},
                "authtype3": {"required": False, "type": "str",
                              "choices": ["pap",
                                          "chap",
                                          "mschap",
                                          "mschapv2"]},
                "auto_dial": {"required": False, "type": "str",
                              "choices": ["enable",
                                          "disable"]},
                "connect_timeout": {"required": False, "type": "int"},
                "dial_cmd1": {"required": False, "type": "str"},
                "dial_cmd2": {"required": False, "type": "str"},
                "dial_cmd3": {"required": False, "type": "str"},
                "dial_on_demand": {"required": False, "type": "str",
                                   "choices": ["enable",
                                               "disable"]},
                "distance": {"required": False, "type": "int"},
                "dont_send_CR1": {"required": False, "type": "str",
                                  "choices": ["enable",
                                              "disable"]},
                "dont_send_CR2": {"required": False, "type": "str",
                                  "choices": ["enable",
                                              "disable"]},
                "dont_send_CR3": {"required": False, "type": "str",
                                  "choices": ["enable",
                                              "disable"]},
                "extra_init1": {"required": False, "type": "str"},
                "extra_init2": {"required": False, "type": "str"},
                "extra_init3": {"required": False, "type": "str"},
                "holddown_timer": {"required": False, "type": "int"},
                "idle_timer": {"required": False, "type": "int"},
                "interface": {"required": False, "type": "str"},
                "lockdown_lac": {"required": False, "type": "str"},
                "mode": {"required": False, "type": "str",
                         "choices": ["standalone",
                                     "redundant"]},
                "network_init": {"required": False, "type": "str"},
                "passwd1": {"required": False, "type": "str"},
                "passwd2": {"required": False, "type": "str"},
                "passwd3": {"required": False, "type": "str"},
                "peer_modem1": {"required": False, "type": "str",
                                "choices": ["generic",
                                            "actiontec",
                                            "ascend_TNT"]},
                "peer_modem2": {"required": False, "type": "str",
                                "choices": ["generic",
                                            "actiontec",
                                            "ascend_TNT"]},
                "peer_modem3": {"required": False, "type": "str",
                                "choices": ["generic",
                                            "actiontec",
                                            "ascend_TNT"]},
                "phone1": {"required": False, "type": "str"},
                "phone2": {"required": False, "type": "str"},
                "phone3": {"required": False, "type": "str"},
                "pin_init": {"required": False, "type": "str"},
                "ppp_echo_request1": {"required": False, "type": "str",
                                      "choices": ["enable",
                                                  "disable"]},
                "ppp_echo_request2": {"required": False, "type": "str",
                                      "choices": ["enable",
                                                  "disable"]},
                "ppp_echo_request3": {"required": False, "type": "str",
                                      "choices": ["enable",
                                                  "disable"]},
                "priority": {"required": False, "type": "int"},
                "redial": {"required": False, "type": "str",
                           "choices": ["none",
                                       "1",
                                       "2",
                                       "3",
                                       "4",
                                       "5",
                                       "6",
                                       "7",
                                       "8",
                                       "9",
                                       "10"]},
                "reset": {"required": False, "type": "int"},
                "status": {"required": False, "type": "str",
                           "choices": ["enable",
                                       "disable"]},
                "traffic_check": {"required": False, "type": "str",
                                  "choices": ["enable",
                                              "disable"]},
                "username1": {"required": False, "type": "str"},
                "username2": {"required": False, "type": "str"},
                "username3": {"required": False, "type": "str"},
                "wireless_port": {"required": False, "type": "int"}

            }
        }
    }

    module = AnsibleModule(argument_spec=fields,
                           supports_check_mode=False)

    # legacy_mode refers to using fortiosapi instead of HTTPAPI
    legacy_mode = 'host' in module.params and module.params['host'] is not None and \
                  'username' in module.params and module.params['username'] is not None and \
                  'password' in module.params and module.params['password'] is not None

    versions_check_result = None
    if not legacy_mode:
        if module._socket_path:
            connection = Connection(module._socket_path)
            fos = FortiOSHandler(connection)

            is_error, has_changed, result = fortios_system(module.params, fos)
            versions_check_result = connection.get_system_version()
        else:
            module.fail_json(**FAIL_SOCKET_MSG)
    else:
        try:
            from fortiosapi import FortiOSAPI
        except ImportError:
            module.fail_json(msg="fortiosapi module is required")

        fos = FortiOSAPI()

        login(module.params, fos)
        is_error, has_changed, result = fortios_system(module.params, fos)
        fos.logout()

    if versions_check_result and versions_check_result['matched'] is False:
        module.warn("Ansible has detected version mismatch between FortOS system and galaxy, see more details by specifying option -vvv")

    if not is_error:
        if versions_check_result and versions_check_result['matched'] is False:
            module.exit_json(changed=has_changed, version_check_warning=versions_check_result, meta=result)
        else:
            module.exit_json(changed=has_changed, meta=result)
    else:
        if versions_check_result and versions_check_result['matched'] is False:
            module.fail_json(msg="Error in repo", version_check_warning=versions_check_result, meta=result)
        else:
            module.fail_json(msg="Error in repo", meta=result)


if __name__ == '__main__':
    main()
